using System;

namespace RTadeusiewicz.NN.NeuralNetworks
{
    /// <summary>
    /// This class represents a linear neuron.
    /// </summary>
    /// <remarks>
    /// A neuron is the basic component of a neural network. It's analogous to the
    /// biological neuron. Its basic property is a set (or a vector, from the
    /// mathematical point of view) of the remembered weights. Each weight is a number
    /// assigned to a particular input, describing how strongly (positively or
    /// negatively) the signal given to that input influences the neuron's response.
    /// </remarks>
    public class Neuron
    {
        /// <summary>
        /// Creates a neuron containing the given number of inputs.
        /// </summary>
        /// <param name="inputCount">The number of neuron's inputs. The weights array
        /// will have the size equal to the given number of inputs.</param>
        public Neuron(int inputCount)
        {
            _weights = new double[inputCount];
            _prevWeights = new double[inputCount];
        }

        // Stores the value of the Weights property.
        private double[] _weights;

        /// Przechowuje poprzednie warto wasnoci Weights.
        private double[] _prevWeights;

        /// <summary>
        /// Returns an array containing the weights of each of this neuron's inputs.
        /// </summary>
        public double[] Weights
        {
            get
            {
                return _weights;
            }
        }

        /// <summary>
        /// Tablica zawierajca poprzednie wagi poszczeglnych wej neuronu.
        /// </summary>
        public double[] PrevWeights
        {
            get
            {
                return _prevWeights;
            }
        }

        /// <summary>
        /// Computes the response of this neuron to the given signal.
        /// <remarks>
        /// To compute the response of a linear network, we need to calculate the
        /// product of each input signal component and its corresponding input weight.
        /// The sum of these products is our response. In practice it turns out that
        /// the more the signal is "similar" to the weights remembered by the neuron,
        /// the more positive the neuron's response will be.
        /// </remarks>
        /// <param name="inputSignals">The signal provided to the neuron inputs. Each
        /// element of the array corresponds to an individual input, so the array's
        /// length should math the number of inputs.</param>
        /// <returns>The neuron's response to the given signal.</returns>
        public virtual double Response(double[] inputSignals)
        {
            // Check if the argument is correct.
            if (inputSignals == null || inputSignals.Length != _weights.Length)
                throw new ArgumentException(
                    "The signal array must have the same length as the weight array.");

            // Calculate the output as the sum of products.
            double result = 0.0;
            for (int i = 0; i < _weights.Length; i++)
                result += _weights[i] * inputSignals[i];
            return result;
        }

        // response neuronu dla exampla 10c white
        public virtual double Response(double[] inputSignals, out double odl)
        {
            // Sprawdzamy, czy argument jest poprawny.
            if (inputSignals == null || inputSignals.Length != _weights.Length)
                throw new ArgumentException(
                    "The signal array must have the same length as the weight array.");

            // Obliczamy wyjcie.
            double result = 0.0;
            odl = 0;
            for (int i = 0; i < _weights.Length; i++)
                odl += Math.Pow(_weights[i] - inputSignals[i], 2);

            result = 1 / (1E-10 + odl);
            return result;
        }

        /// <summary>
        /// Oblicza moc podanego sygnau, uywajc zadanej miary.
        /// </summary>
        /// <param name="signals">Sygna, ktrego moc obliczamy</param>
        /// <param name="norm">Norma, jak mierzymy moc (euklidesowa lub manhattaska)
        /// </param>
        /// <returns>Obliczona moc</returns>
        /// <remarks>
        /// Moc sygnau wpywa na to, jak bardzo "istotny" jest on dla neuronu.
        /// Jeli zamiast sygnau podamy wartoci wspczynnikw wag, dostaniemy
        /// moc ladu pamiciowego - czyli warto okrelajc, jak bardzo
        /// "zdecydowany" jest dany neuron.
        /// </remarks>
        /// <seealso cref="MemoryTraceStrength(StrengthNorm)"/>
        public static double Strength(double[] signals, StrengthNorm norm)
        {
            double strength = 0;
            switch (norm)
            {
                case StrengthNorm.Manhattan:
                    foreach (double s in signals)
                        strength += Math.Abs(s);
                    return strength;

                case StrengthNorm.Euclidean:
                    foreach (double s in signals)
                        strength += s * s;
                    return Math.Sqrt(strength);

                default:
                    return 0.0;
            }
        }

        /// <summary>
        /// Oblicza moc ladu pamiciowego.
        /// </summary>
        /// <param name="norm">Norma, jak mierzymy moc (euklidesowa lub manhattaska)
        /// </param>
        /// <returns>Obliczona moc ladu pamiciowego</returns>
        /// <remarks>
        /// Moc ladu pamiciowego okrela, jak bardzo "zdecydowany" jest
        /// neuron. Im wiksza moc ladu pamiciowego, tym bardziej radykalne bd
        /// jego "opinie".
        /// </remarks>
        public double MemoryTraceStrength(StrengthNorm norm)
        {
            return Strength(_weights, norm);
        }

        /// <summary>
        /// Normalizuje podany sygna.
        /// </summary>
        /// <param name="signals">Sygna do normalizacji. Metoda zmienia zawarto
        /// tej tablicy, wic aby zachowa orygina, trzeba go najpierw gdzie
        /// skopiowa.</param>
        /// <remarks>
        /// Normalizacja polega na tym, e wszystkie skadowe s zmieniane tak, aby
        /// nie ulegy zmianie ich proporcje, za to moc sygnau wyniosa dokadnie 1.
        /// Innymi sowy, wektor sygnau jest skalowany tak, aby mia dugo 1 (wg
        /// normy Euklidesowej).
        /// </remarks>
        public static void Normalize(double[] signals)
        {
            double strength = Strength(signals, StrengthNorm.Euclidean);
            for (int i = 0; i < signals.Length; i++)
                signals[i] /= strength;
        }

        /// <summary>
        /// Uczy neuron reakcji na dany sygna. Jest to "uczenie z nauczycielem".
        /// </summary>
        /// <param name="signals">Sygna, na ktry neuron ma zareagowa.</param>
        /// <param name="expectedOutput">Prawidowa odpowied neuronu.</param>
        /// <param name="ratio">Wspczynnik uczenia. Im wikszy, tym szybszy bdzie
        /// proces uczenia, lecz mniejsza jego stabilno.</param>
        /// <param name="previousResponse">Parametr wyjciowy; metoda umieszcza
        /// w nim odpowied neuronu na zadany sygna przed uczeniem.</param>
        /// <param name="previousError">Parametr wyjciowy; metoda umieszcza w nim
        /// wielko bdu popenianego przez neuron przed uczeniem.</param>
        public void Learn(double[] signals, double expectedOutput, double ratio,
            out double previousResponse, out double previousError)
        {
            /* Nie sprawdzamy rozmiaru tablicy sygnaw - Response() zrobi to
             * za nas. */
            previousResponse = Response(signals);
            previousError = expectedOutput - previousResponse;
            for (int i = 0; i < _weights.Length; i++)
                _weights[i] += ratio * previousError * signals[i];
        }

        public void Learn(double[] signals, double[] previous_weights, double error, double sigma, double ratio, double momentum)
        {
            for (int i = 0; i < _weights.Length; i++)
                _weights[i] += ratio * sigma * signals[i] - momentum * (_weights[i] - previous_weights[i]);
        }

        public void Randomize(Random randomGenerator, double min, double max)
        {
            double length = max - min;
            for (int i = 0; i < _weights.Length; i++)
                _weights[i] = min + length * randomGenerator.NextDouble();
        }

        /* dodane dla przykladu 10a,ax,b */
        public void Learn(double[] signals, double etha, double max)
        {

            double previous_response = Response(signals);

            if (previous_response < 0.2 * max)
                previous_response *= 0.3;
            if (previous_response < 0)
                previous_response *= 0.1;

            for (int i = 0; i < _weights.Length; i++)
            {
                _prevWeights[i] = _weights[i];
                _weights[i] += etha * previous_response * (signals[i] - _weights[i]);
            }
        }

        /* dodane dla przykladu 10c samouczenie z konkurencja */
        public void LearnSelf(double[] signals, double etha)
        {
            for (int i = 0; i < _weights.Length; i++)
            {
                _prevWeights[i] = _weights[i];
                _weights[i] += etha * (signals[i] - _weights[i]);
            }
        }

        public void Randomize(Random randomGenerator, double min, double max, double epsilon)
        {
            double length = max - min;
            for (int i = 0; i < _weights.Length; i++)
            {
                _weights[i] = min + length * randomGenerator.NextDouble();
                if (Math.Abs(_weights[i]) < epsilon)
                    _weights[i] = epsilon;
            }
        }

    }

    /// <summary>
    /// Rodzaje miar mocy sygnau.
    /// </summary>
    public enum StrengthNorm
    { 

        /// <summary>
        /// Moc obliczana jako suma wartoci bezwzgldnych poszczeglnych skadnikw
        /// sygnau. Jej nazwa wzia si std, e moe ona suy do okrelania
        /// dugoci, jak naley przejecha w wielkim miecie midzy dwoma punktami,
        /// ale nie w linii prostej, tylko po prostopadych do siebie ulicach.
        /// </summary>
        Manhattan,

        /// <summary>
        /// Moc obliczana jako pierwiastek kwadratowy sumy kwadratw poszczeglnych
        /// skadnikw sygnau. Innymi sowy - jest to odlego w przestrzeni midzy
        /// pocztkiem ukadu wsprzdnych a punktem o wsprzdnych odpowiadajcych
        /// skadnikom sygnau.
        /// </summary>
        Euclidean
    }
}
